;************************************************************
;*****	           Pulzn mi ESR             	*****
;************************************************************

	list p=16f873

; Include files, change directory if needed
	#include <p16f873.inc>		; processor specific variable definitions
	#include <dev_fam.inc>		; definitions for MATH routines
	#include <math16.inc>		; definitions for MATH routines
	
	__CONFIG _CP_OFF & _WDT_OFF & _BODEN_OFF & _PWRTE_ON & _XT_OSC & _WRT_ENABLE_ON & _LVP_OFF & _DEBUG_ON & _CPD_OFF 


; Init values
iOPTION_REGISTER	equ	B'00000011';		     TMR0 rate	    WDT rate
;				  +++  PS2..PS0	000	1:2		1:1
;				   +- Prescaler	001	1:4		1:2
;				  			010	1:8		1:4		
;				  			011	1:16		1:8
;				  			100	1:32		1:16
;				  			101	1:64		1:32
;				  			110	1:128		1:64
;				  			111	1:256		1:128
;				  +--- PSA		Prescaler assigment bit  0= Timer0	1= WDT
;				  +---- T0SE		TMR0 Select Edge on RA4  0= L->H	1= H->L
;				  +----- T0CS		TMR0 Clock Select	 0= inter.	1= ext.  RA4 / T0CKI
;				  +------ INTEDG	Interrupt edge select	 0= H->L	1= L->H  RB0 / INT	 
;				  +------- RBPU		PORT pull up		 0= enabled	1=disabled

iINTCON			equ	B'10100000'
;				  + RBIF		Port B change interrupt flag bit - only RB7..RB4 - software must clear
;				  +- INTF		RB0 / INT external interrupt flag bit		 - software must clear
;				  +-- T0IF		Timer 0 interrupt flag bit			 - software must clear
;				  +--- RBIE		RB7..RB4 Port Change interrupt Enable bit   0 - disabled   1 - enabled
;				  +---- INTE		RB0 / INT  Change interrupt Enable bit      0 - disabled   1 - enabled
;				  +----- T0IE		TMR0 Overflow interrupt enable		    0 - disabled   1 - enabled
;				  +------ PEIE		Peripheral interrupt enable bit		    0 - disabled   1 - enabled
;				  +------- GIE		Global interrupt enable bit		    0 - disabled   1 - enabled

iPIE1			equ	B'00000000'
;				  + TMR1IE	0= Int.disabled	1= Int.enabled (Timer 1 overflow)
;				  +- TMR2IE	0= Int.disabled	1= Int.enabled (TMR2 to PR2 match)
;				  +-- CCP1IE	0= Int.disabled	1= Int.enabled
;				  +--- SSPIE	0= Int.disabled	1= Int.enabled
;				  +---- TxIE		0= Int.disabled	1= Int.enabled
;				  +----- RcIE		0= Int.disabled	1= Int.enabled
;				  +------ ADIE		0= Int.disabled	1= Int.enabled
;				  +------- PSPIE	0= Int.disabled	1= Int.enabled

iTRISA			equ	B'00001001'
;				  + RA0		0= output	1= input
;				  +- RA1		0= output	1= input
;				  +-- RA2		0= output	1= input
;				  +--- RA3		0= output	1= input
;				  +---- RA4		0= output	1= input
;				  +----- RA5		0= output	1= input
;				  +------ RA6		doesn't exist
;				  +------- RA7		doesn't exist

iTRISB			equ	B'11000000'
;				  + RB0		0= output	1= input
;				  +- RB1		0= output	1= input
;				  +-- RB2		0= output	1= input
;				  +--- RB3		0= output	1= input
;				  +---- RB4		0= output	1= input
;				  +----- RB5		0= output	1= input
;				  +------ RB6		0= output	1= input
;				  +------- RB7		0= output	1= input

iTRISC			equ	B'00000000'
;				  + RC0		0= output	1= input
;				  +- RC1		0= output	1= input
;				  +-- RC2		0= output	1= input
;				  +--- RC3		0= output	1= input
;				  +---- RC4		0= output	1= input
;				  +----- RC5		0= output	1= input
;				  +------ RC6		0= output	1= input
;				  +------- RC7		0= output	1= input

iADCON0			equ	B'01000001'
;				  + ADON		0= AD disabled	1= AD enabled
;				  +- none		unimplemented
;				  +-- GO/DONE	0= AD is ready	1= AD is running
;				  +++--- CHS2..0	000 = Channel 0
;				  			001 = Channel 1
;				  			010 = Channel 2
;				  			011 = Channel 3
;				  			100 = Channel 4
;				  			101 = Channel 5
;				  			110 = Channel 6
;				  			111 = Channel 7
;				  ++------ ADCS1..0	00  = Fosc/2
;				  			01  = Foasc/8
;				  			10  = Foasc/32
;				  			11  = RC

iADCON1			equ	B'10000101'		;	  RE2 RE1 RE0 RA5 RA3 RA2 RA1 RA0
;				  ++++ PCFG3..0	0000	=  A   A   A   A   A   A   A   A
;				  			0001	=  A   A   A   A   V+  A   A   A
;				  			0010    =  D   D   D   A   A   A   A   A
;				  			0011    =  D   D   D   A   V+  A   A   A  
;				  			0100	=  D   D   D   D   A   D   A   A
;				  			0101	=  D   D   D   D   V+  D   A   A
;				  			0110	=  D   D   D   D   D   D   D   D
;				  			0111	=  D   D   D   D   D   D   D   D
;				  			1000	=  A   A   A   A   V+  V-  A   A
;				  			1001	=  D   D   A   A   A   A   A   A
;				  			1010	=  D   D   A   A   V+  A   A   A
;				  			1011	=  D   D   A   A   V+  V-  A   A
;				  			1100	=  D   D   D   A   V+  V-  A   A
;				  			1101	=  D   D   D   D   V+  V-  A   A
;				  			1110	=  D   D   D   D   D   D   D   A
;				  			1111	=  D   D   D   D   V+  V-  D   A
;				  +---- none		unimplemented
;				  +----- none		unimplemented
;				  +------ none		unimplemented
;				  +------- ADFM		0= left just.	1= Right justified


;	Variable definition in register Bank 0

TickCNT	equ	37h		;Tikcs counter
Second	equ	38h		;Second counter
WREGtmp	equ	39h		;Temporary WREG
				
Disp100	equ	3Ah		;DispCNT=3
Disp_10	equ	3Bh		;DispCNT=2
Disp__1	equ	3Ch		;DispCNT=1
DispCNT	equ	3Dh		;Counter of displayed place

ADlow	equ	3Eh		;AD result - low byte
ADhigh	equ	3Fh		;AD result - high byte

				;Voltage on the output amplifier
AMPlow	equ	40h		;low byte
AMPhigh	equ	41h		;high byte
ADstat	equ	42h		;AD status
ADdone	equ	0x00		;=1 -> read ADhigh & ADlow

WREGbck	equ	70h		;backup WREG
STATbck	equ	71h		;backup STATUS

;	Variable definition in register Bank 1
WREGbc1	equ	70h+80h		;backup WREG in Bank 1



;		Bits definition

iTim0	equ	0x02		; Timer 0 interrupt flag bit in INTCON register


IntPsec equ	D'244'	;nuber of interrupts per second
iDispCNT EQU	D'3'	;number of digits of the display




	; Start at the reset vector

	org	0x0000
	pagesel	Start
	goto	Start

	org	0x0004
Intterupt_vector
	pagesel	Interrupt_service_routine
	goto	Interrupt_service_routine

Start
;************************************
;*****	Inicialization		*****
;************************************
	banksel	PORTA
	movlw	0xFF
	movwf	PORTA
	movwf	PORTB
	movwf	PORTC

	movlw	IntPsec		;Init ticks counter
	movwf	TickCNT

	movlw	1		;Init display variables
	movwf	DispCNT		;Init counter of displayed place
	movlw	~B'00111111'
	movwf	Disp__1		;Init display registers
	movwf	Disp_10
	movwf	Disp100

	banksel	OPTION_REG	;Bank 1
	movlw	iOPTION_REGISTER
	movwf	OPTION_REG
	movlw	iTRISA
	movwf	TRISA		;Port A configuration
	movlw	iTRISB
	movwf	TRISB		;Port B configuration
	movlw	iTRISC
	movwf	TRISC		;Port C configuration
	movlw	iPIE1
	movwf	PIE1		;Init Peripheral Interrupt mask register 1
	movlw	iINTCON
	movwf	INTCON		;Init Interrupt configuration register
	movlw	iADCON1		;Load configuration of analog/digitals inputs
	movwf	ADCON1		;Set up AD


	banksel	PORTA		;Bank 0
	movlw	0x80
	movwf	PORTC
	movlw	0xFE
	movwf	PORTB
	movlw	iADCON0		;Load configuration of AD convertor
	movwf	ADCON0		;Set up AD

	bsf	ADCON0,GO_DONE	;Start AD/conversion	
	banksel	ADCON0
Loop_Start
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Loop_Start
	nop
	nop
	nop

Low_Range			;***** LOWEST CURRENT PULSE ****
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Low_Range
	nop
	nop
	nop
Loop_Low
	bsf	ADCON0,GO_DONE	;Start AD/conversion	
Loop_Low_0
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Loop_Low_0	;			yes -> wait (goto Loop)
				;Read AD conversion result
	banksel	ADRESL		;Select register bank 1
	movf	ADRESL,W	;read outpout voltage of the amplifier (without current impuls)
	banksel	ADRESH		;Select register bank0
	movwf	AMPlow
	movf	ADRESH,W
	movwf	AMPhigh
;				 Apply CURRENT impulse
	bcf	PORTA,0x2	;Switch off transistor T9
	bcf	PORTB,0x3	;Current impuls applied
	nop
	nop	
	bsf	ADCON0,GO_DONE	;Start AD/conversion
	nop	
	nop	
	nop	
	nop	
	nop	
	nop	
	nop
	nop	
	bsf	PORTB,0x3	;end of current impuls
	bsf	PORTA,0x2	;Swich on transistor T9
Loop_Low_1
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Loop_Low_1	;			yes -> wait (goto Loop_Low_1)
				;Read AD conversion result
	banksel	ADRESL		;Select register bank1
	movf	ADRESL,W
	banksel	ADRESH
	movwf	AARGB2		;low byte of AD conversion
	movf	ADRESH,W
	movwf	AARGB1		;high byte of AD conversion
	clrw			;WREG = 0
	movwf	AARGB0		;highest byte of 24bit format is  always 0
				;AARGB0..2 = data for conversion 24bit -> float
	movlw	-0x03		;Check correct range
	addwf	AARGB1,W	;Is higher byte of ADC < 0x3 ?
	btfss	STATUS,C	;                             yes -> goto Wait_Low_12, check minimum
	goto	Wait_Low_12
	btfss	STATUS,Z
	goto	DisplayOverFlow ; Display Overflow
	movlw	-0x18
	addwf	AARGB2,W
	btfsc	STATUS,C	;                             yes -> goto Wait_12
	goto	DisplayOverFlow ; Display Overflow
Wait_Low_12
	movlw	0x20
	movwf	AARGB0		;Delay = 32 tacts = 32us
Wait_Low_2				;Wait
	decfsz	AARGB0,F
	goto	Wait_Low_2
	bsf	ADCON0,GO_DONE	;Start A/D conversion	
Loop_Low_2
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Loop_Low_2	;			yes -> wait (goto Loop_2)
				;***** Compute result *****
	banksel	ADRESL		;Select register bank 1
	movf	ADRESL,W	;read outpout voltage of the amplifier (without current impuls)
	banksel	ADRESH		;Select register bank 0
	addwf	AMPlow,F
	btfsc	STATUS,C
	incf	AMPhigh,F
	movf	ADRESH,W
	addwf	AMPhigh,F

	bcf	STATUS,C	;unsigned shift left -> unsigned / 2
	rrf	AMPhigh,F
	rrf	AMPlow,F
	btfsc	STATUS,C
	incf	AMPlow,F	;AMP = output voltage of the amplifier without current impulse
				;This is mean value before and after applying impulse
	movf	AMPlow,W	;AARGB0..2 = Impulse voltage - mean value of the output offset voltage
	subwf	AARGB2,F
	btfss	STATUS,C
	incf	AMPhigh,F
	movf	AMPhigh,W	
	subwf	AARGB1,F

	movf	AARGB1,W	; move highest byte of result into WREG
	btfss	STATUS,Z	; Is AARGB1 = 0 ?
	goto	Wait_Low_13	;		no -> goto Wait_Low_13
	movlw	-0x50		;
	addwf	AARGB2,W	; Check AARGB2
	btfss	STATUS,C	; Is AARGB2 < 0x50 ?
	goto	Middle_Range	;		no -> use bigger current
Wait_Low_13

	call	FLO32		;convert it to float
	movlw	0x7F		;BARG = conversion constant
	movwf	BEXP
	movlw	0x00
	movwf	BARGB0
	movlw	0x00
	movwf	BARGB1
	movlw	0x00
	movwf	BARGB2
	call	FPM32		;AARG<-AARG * BARG
	call	INT32		;Int(AARG)->AARG
	movf	AARGB1,W
	movwf	AARGB0		;AARG1 -> AARG0
	movf	AARGB2,W
	movwf	AARGB1		;AARG2 -> AARG1
	movlw	D'10'
	movwf	BARGB0		;   10 -> BARG0
	call	FXD1608U
	movf	REMB0,W
	call	BCDtoDISP
	movwf	Disp__1
	call	FXD1608U
	movf	REMB0,W
	call	BCDtoDISP
	movwf	Disp_10
	call	FXD1608U
	movf	REMB0,W
	call	BCDtoDISP
	movwf	Disp100
	bcf	Disp100,0x07
Display_Low
	movf	Second,W
	movwf	LOOPCNT
WaitForSecond_Low
	movf	LOOPCNT,W
	subwf	Second,W
	btfsc	STATUS,Z
	goto	WaitForSecond_Low
	goto	Loop_Low
DisplayOverFlow
	movlw	~B'00000001'	;Upper line
	movwf	Disp100
	movwf	Disp_10
	movwf	Disp__1
	goto	Display_Low	; Display upperlines = overflow

Middle_Range			;***** MIDDLE CURRENT PULSE *****
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Middle_Range
	nop
	nop
	nop
Loop_Middle
	bsf	ADCON0,GO_DONE	;Start AD/conversion
Loop_Middle_0
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Loop_Middle_0	;			yes -> wait (goto Loop)
				;Read AD conversion result
	banksel	ADRESL		;Select register bank 1
	movf	ADRESL,W	;read outpout voltage of the amplifier (without current impuls)
	banksel	ADRESH		;Select register bank0
	movwf	AMPlow
	movf	ADRESH,W
	movwf	AMPhigh
;				 Apply CURRENT impulse
	bcf	PORTA,0x2	;Switch off transistor T9
	bcf	PORTB,0x4	;Current impuls applied
	nop
	nop	
	bsf	ADCON0,GO_DONE	;Start AD/conversion
	nop	
	nop	
	nop	
	nop	
	nop	
	nop	
	nop
	nop
	nop
	bsf	PORTB,0x4	;end of current impuls
	bsf	PORTA,0x2	;Swich on transistor T9

Loop_Middle_1
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Loop_Middle_1	;			yes -> wait (goto Loop_Middle_1)
				;Read AD conversion result
	banksel	ADRESL		;Select register bank1
	movf	ADRESL,W
	banksel	ADRESH
	movwf	AARGB2		;low byte of AD conversion
	movf	ADRESH,W
	movwf	AARGB1		;high byte of AD conversion
	clrw			;WREG = 0
	movwf	AARGB0		;highest byte of 24bit format is  always 0
				;AARGB0..2 data for conversion 24bit -> float
	movlw	-0x03		;Check correct range
	addwf	AARGB1,W	;Is higher byte of ADC < 0x3 ?
	btfss	STATUS,C	;                             yes -> goto Wait_Middle_12
	goto	Wait_Middle_12
	btfss	STATUS,Z
	goto	Low_Range	; use lower current
	movlw	-0x18
	addwf	AARGB2,W
	btfsc	STATUS,C	;                             yes -> goto Wait_Middle_12
	goto	Low_Range	; use lower current
Wait_Middle_12
	movlw	0x20
	movwf	AARGB0		;Delay = 32 tacts = 32us
Wait_Middle_2				;Wait
	decfsz	AARGB0,F
	goto	Wait_Middle_2
	bsf	ADCON0,GO_DONE	;Start A/D conversion	
Loop_Middle_2
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Loop_Middle_2	;			yes -> wait (goto Loop_2)
				;***** Compute result *****
	banksel	ADRESL		;Select register bank 1
	movf	ADRESL,W	;read outpout voltage of the amplifier (without current impuls)
	banksel	ADRESH		;Select register bank 0
	addwf	AMPlow,F
	btfsc	STATUS,C
	incf	AMPhigh,F
	movf	ADRESH,W
	addwf	AMPhigh,F
	bcf	STATUS,C	;unsigned shift left -> unsigned / 2
	rrf	AMPhigh,F
	rrf	AMPlow,F
	btfsc	STATUS,C
	incf	AMPlow,F	;AMP = output voltage of the amplifier without current impulse
				;This is mean value before and after applying impulse
	movf	AMPlow,W	;AARGB0..2 = Impulse voltage - mean value of the output offset voltage
	subwf	AARGB2,F
	btfss	STATUS,C
	incf	AMPhigh,F
	movf	AMPhigh,W	
	subwf	AARGB1,F

	movf	AARGB1,W	; move highest byte of result into WREG
	btfss	STATUS,Z	; Is AARGB1 = 0 ?
	goto	Wait_Middle_13	;		no -> goto Wait_Low_13
	movlw	-0x50		;
	addwf	AARGB2,W	; Check AARGB2
	btfss	STATUS,C	; Is AARGB2 < 0x30 ?
	goto	Loop_High	;		no -> use bigger current
Wait_Middle_13
	call	FLO32		;convert it to float
	movlw	0x7F		;BARG = conversion constant
	movwf	BEXP
	movlw	0x00
	movwf	BARGB0
	movlw	0x00
	movwf	BARGB1
	movlw	0x00
	movwf	BARGB2
	call	FPM32		;AARG<-AARG * BARG
	call	INT32		;Int(AARG)->AARG
	movf	AARGB1,W
	movwf	AARGB0		;AARG1 -> AARG0
	movf	AARGB2,W
	movwf	AARGB1		;AARG2 -> AARG1
	movlw	D'10'
	movwf	BARGB0		;   10 -> BARG0
	call	FXD1608U
	movf	REMB0,W
	call	BCDtoDISP
	movwf	Disp__1
	call	FXD1608U
	movf	REMB0,W
	call	BCDtoDISP
	movwf	Disp_10
	call	FXD1608U
	movf	REMB0,W
	call	BCDtoDISP
	movwf	Disp100
	bcf	Disp100,0x07
	movf	Second,W
	movwf	LOOPCNT
WaitForSecond_Middle
	movf	LOOPCNT,W
	subwf	Second,W
	btfsc	STATUS,Z
	goto	WaitForSecond_Middle
	goto	Loop_Middle

Loop_High			;***** HIGHEST CURRENT PULSE *****
	bsf	ADCON0,GO_DONE	;Start AD/conversion	
Loop_High_0
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Loop_High_0	;			yes -> wait (goto Loop)
				;Read AD conversion result
	banksel	ADRESL		;Select register bank 1
	movf	ADRESL,W	;read outpout voltage of the amplifier (without current impuls)
	banksel	ADRESH		;Select register bank0
	movwf	AMPlow
	movf	ADRESH,W
	movwf	AMPhigh
;				 Apply CURRENT impulse
	bcf	PORTA,0x2	;Switch off transistor T9
	bcf	PORTB,0x5	;Current impuls applied
	nop
	nop	
	bsf	ADCON0,GO_DONE	;Start AD/conversion
	nop	
	nop	
	nop	
	nop	
	nop	
	nop	
	nop
	nop	
	bsf	PORTB,0x5	;end of current impuls
	bsf	PORTA,0x2	;Swich on transistor T9

Loop_High_1
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Loop_High_1	;			yes -> wait (goto Loop_Middle_1)
				;Read AD conversion result
	banksel	ADRESL		;Select register bank1
	movf	ADRESL,W
	banksel	ADRESH
	movwf	AARGB2		;low byte of AD conversion
	movf	ADRESH,W
	movwf	AARGB1		;high byte of AD conversion
	clrw			;WREG = 0
	movwf	AARGB0		;highest byte of 24bit format is  always 0
				;AARGB0..2 data for conversion 24bit -> float
	movlw	-0x03		;Check correct range
	addwf	AARGB1,W	;Is higher byte of ADC < 0x3 ?
	btfss	STATUS,C	;                             yes -> goto Wait_12
	goto	Wait_High_12
	btfss	STATUS,Z
	goto	Middle_Range
	movlw	-0x18
	addwf	AARGB2,W
	btfsc	STATUS,C	;                             yes -> goto Wait_12
	goto	Middle_Range
Wait_High_12
	movlw	0x20
	movwf	AARGB0		;Delay = 32 tacts = 32us
Wait_High_2				;Wait
	decfsz	AARGB0,F
	goto	Wait_High_2
	bsf	ADCON0,GO_DONE	;Start A/D conversion	
Loop_High_2
	btfsc	ADCON0,GO_DONE	;Is A/D conversion running?
	goto	Loop_High_2	;			yes -> wait (goto Loop_2)
				;***** Compute result *****
	banksel	ADRESL		;Select register bank 1
	movf	ADRESL,W	;read outpout voltage of the amplifier (without current impuls)
	banksel	ADRESH		;Select register bank 0
	addwf	AMPlow,F
	btfsc	STATUS,C
	incf	AMPhigh,F
	movf	ADRESH,W
	addwf	AMPhigh,F

	bcf	STATUS,C	;unsigned shift left -> unsigned / 2
	rrf	AMPhigh,F
	rrf	AMPlow,F
	btfsc	STATUS,C
	incf	AMPlow,F	;AMP = output voltage of the amplifier without current impulse
				;This is mean value before and after applying impulse
	movf	AMPlow,W	;AARGB0..2 = Impulse voltage - mean value of the output offset voltage
	subwf	AARGB2,F
	btfss	STATUS,C
	incf	AMPhigh,F
	movf	AMPhigh,W	
	subwf	AARGB1,F
	call	FLO32		;convert it to float
	movlw	0x7F		;BARG = conversion constant
	movwf	BEXP
	movlw	0x00
	movwf	BARGB0
	movlw	0x00
	movwf	BARGB1
	movlw	0x00
	movwf	BARGB2
	call	FPM32		;AARG<-AARG * BARG
	call	INT32		;Int(AARG)->AARG
	movf	AARGB1,W
	movwf	AARGB0		;AARG1 -> AARG0
	movf	AARGB2,W
	movwf	AARGB1		;AARG2 -> AARG1
	movlw	D'10'
	movwf	BARGB0		;   10 -> BARG0
	call	FXD1608U
	movf	REMB0,W
	call	BCDtoDISP
	movwf	Disp__1
	call	FXD1608U
	movf	REMB0,W
	call	BCDtoDISP
	movwf	Disp_10
	call	FXD1608U
	movf	REMB0,W
	call	BCDtoDISP
	movwf	Disp100
	bcf	Disp100,0x07
	movf	Second,W
	movwf	LOOPCNT
WaitForSecond_High
	movf	LOOPCNT,W
	subwf	Second,W
	btfsc	STATUS,Z
	goto	WaitForSecond_High
	goto	Loop_High


;-----------------------------------------------------------------------------------------------------		
;					Interrupt_service_routine
;-----------------------------------------------------------------------------------------------------		
;Input:		-
;Output:	-
;USE:		Register bank 0
;Change:	WREGbck, WREGbc1, STATbck
Interrupt_service_routine	;push WREG and STATUS
	movwf	WREGbck		;save content of working register, doesn't matter on current bank
	swapf	STATUS,W	;swap nibbles of the STATUS and save content into WREG
	bcf	STATUS,RP0	;force Bank 0 (original content of the STATUS is already saved in WREG)
	movwf	STATbck		;save content of the WREG (=swapped STATUS)

				;Start Interrupt Service Rutine
	btfss	INTCON,T0IF	;Was interrupt invoked by Timer0 ?
	goto	Check_PIR1	; 			no -> check other interrupts
	btfsc	PORTA,0x05
	goto	Clear_Bit5_PORTA
	bsf	PORTA,0x05
	goto	Timer_0_Interrupt_routine
Clear_Bit5_PORTA
	bcf	PORTA,0x05
Timer_0_Interrupt_routine
	decfsz	TickCNT,F
	goto	Timer_0_Cont1
	movlw	IntPsec
	movwf	TickCNT
	incf	Second,F
Timer_0_Cont1
	call	BlankCode	;w=Blank code for LED display
	banksel	PORTC		;Select Bank 0
	movwf	PORTC		;Switch off display
	movf	DispCNT,W	;Get number of digit
	btfsc	STATUS,Z	;Is place = 0 ? no -> ignore next instruction
	goto	Timer_0_Read_Key; goto read keys
	addlw	-.3		;Is DispCNT = 3 ?
	btfsc	STATUS,C	;		  no -> ignore next instruction
	goto	Timer_0_DispCNT_3;		 yes -> goto Timer_0_DispCNT_3
	addlw	D'1'		;Is DispCNT = 2?
	btfsc	STATUS,C	;		  no -> ignore next instruction
	goto	Timer_0_DispCNT_2;		 yes -> goto Timer_0_DispCNT_2
Timer_0_DispCNT_1
	bsf	PORTB,2
	bsf	PORTB,1
	bcf	PORTB,0
	movf	Disp__1,W	;load
	movwf	PORTC
	decf	DispCNT,F
	goto	Timer_0_End_of_Service
Timer_0_DispCNT_2
	bsf	PORTB,2
	bcf	PORTB,1
	bsf	PORTB,0
	movf	Disp_10,W	;load
	movwf	PORTC

	decf	DispCNT,F
	goto	Timer_0_End_of_Service
Timer_0_DispCNT_3
	bcf	PORTB,2
	bsf	PORTB,1
	bsf	PORTB,0
	movf	Disp100,W	;load
	movwf	PORTC
	decf	DispCNT,F
	goto	Timer_0_End_of_Service
Timer_0_Read_Key
	movlw	iDispCNT	;Load init value of the counter of digits
	movwf	DispCNT		;Save it

Timer_0_End_of_Service
	bcf	INTCON, T0IF	;Clear T0IF interrupt flag
	goto	Check_PIR1	;Check other interrupts

Check_PIR1			;Any interrupt in PIR1 ?
	movf	PIR1,W		;Load PIR1
	btfsc	STATUS,Z	;Any interrupt? (PIR1 = 0)
	goto	End_of_Interrupt;                      no -> End_of_Interrupt
				;Search for interrupt identity
	btfss	PIR1,ADIF	;Is it A/D convertor ?
	goto	Other_PIR1	;		no -> check other interrupts
	bcf	PIR1,ADIF	;Clear interrupt flag of A/D convertor
	banksel	ADRESL		;Select Bank for ADRESL (Bank 1)
	movf	ADRESL,W	;Read AD conversion result - low byte
	banksel	ADlow		;Select Bank for ADlow (Bank 0)
	movwf	ADlow		;save result to memory - low byte
	movf	ADRESH,W	;Read AD conversion result - high byte
	movwf	ADhigh		;save result to memory - high byte
	bsf	ADstat,ADdone	;Result in ADhigh & ADlow is correct
Other_PIR1
				;End of Interrupt Service Routine
End_of_Interrupt		;Restore STATUS and WREG
	swapf	STATbck,W	;swap nibles of the STATUS and load into WREG
	movwf	STATUS		;restore old STATUS include register bank
	swapf	WREGbck,F	;swap nibles of the backuped WREG
	swapf	WREGbck,W	;restore content of the WREG
	retfie			;Return from interrupt


;-----------------------------------------------------------------------------------------------------		
;					 Unsigned 8bit divide
;-----------------------------------------------------------------------------------------------------		
;Input: AARGB0	= divide
;	BARGB0	= divider
;Output:w	=
;	REMB0	= reminder
;Change:WREG, STATUS, AARGB0, BARGB0, REMB0, LOOPCNT
;       Max Timing: 2+7*12+11 = 97 clks
;       Min Timing: 2+7*11+10 = 89 clks
;       PM: 13     DM: 4
UDIV0808
	clrf	REMB0
	clrf	STATUS
	movlw	8
	movwf	LOOPCNT
LOOPU0808A
	rlf	AARGB0,W
	rlf	REMB0, F
	movf	BARGB0,W
	subwf	REMB0, F
	btfsc	STATUS,C
	goto	UOK88A          
	addwf	REMB0, F
	bcf	STATUS,C
UOK88A
	rlf	AARGB0, F
	decfsz	LOOPCNT, F
	goto	LOOPU0808A
	return

;-----------------------------------------------------------------------------------------------------		
;					BCD -> Display code
;-----------------------------------------------------------------------------------------------------		
;Input: w	= value for translation (low nibble only is important)
;Output:w	= result
;Change:w, STATUS, PCLATCH, WREGtmp
	org	0x0200
BCDtoDISP
	andlw	0x0F		;only low nibble
	movwf	WREGtmp		;save parameter
				;=pageselw BCDtoDISP
	movlw	high (BCDtoDISP);setup correct pate in PCLATCH
	movwf	PCLATH

	movf	WREGtmp,W	;restore parameter
	addwf	PCL,F		;compute jump & jump
	retlw	~B'00111111'	;0
	retlw	~B'00000110'	;1
	retlw	~B'01011011'	;2
	retlw	~B'01001111'	;3
	retlw	~B'01100110'	;4
	retlw	~B'01101101'	;5
	retlw	~B'01111101'	;6
	retlw	~B'00000111'	;7
	retlw	~B'01111111'	;8
	retlw	~B'01100111'	;9
	retlw	~B'00000000'	;A -> blank
	retlw	~B'00000000'	;B -> blank
	retlw	~B'00000000'	;C -> blank
	retlw	~B'00000000'	;D -> blank
	retlw	~B'00000000'	;E -> blank
	retlw	~B'01000000'	;F -> minus
;-----------------------------------------------------------------------------------------------------		
;					Return black code
;-----------------------------------------------------------------------------------------------------		
;Input:	-
;Output:w	= black code
;Change:w

BlankCode
	retlw	~B'00000000'	;return blank code

;	RCS Header $Id: fp32.a16 2.8 1996/10/07 13:50:59 F.J.Testa Exp $	$Revision: 2.8 $
;       PIC16 32 BIT FLOATING POINT LIBRARY
;
;       Unary operations: both input and output are in AEXP,AARG
;
;       Binary operations: input in AEXP,AARG and BEXP,BARG with output in AEXP,AARG
;
;       All routines return WREG = 0x00 for successful completion, and WREG = 0xFF
;       for an error condition specified in FPFLAGS.
;
;       All timings are worst case cycle counts
;
;         Routine               Function
;
;       FLO2432         24 bit integer to 32 bit floating point conversion
;       FLO32
;
;               Timing:            RND
;                                0       1
;                       0       104     104
;                  SAT
;                       1       110     110
;
;       NRM3232   32 bit normalization of unnormalized 32 bit floating point numbers
;       NRM32
;
;               Timing:            RND
;                               0       1
;                       0       90      90
;                  SAT
;                       1       96      96
;
;
;       INT3224         32 bit floating point to 24 bit integer conversion
;       INT32
;
;
;               Timing:            RND
;                                0       1
;                       0       104      112
;                  SAT
;                       1       104      114
;
;       FLO3232 32 bit integer to 32 bit floating point conversion
;
;               Timing:            RND
;                               0       1
;
;                       0       129     145
;                  SAT
;                       1       129     152
;
;       NRM4032 32 bit normalization of unnormalized 40 bit floating point numbers
;
;               Timing:            RND
;                               0       1
;
;                       0       112     128
;                  SAT
;                       1       112     135
;
;
;       INT3232         32 bit floating point to 32 bit integer conversion
;
;
;               Timing:            RND
;                               0       1
;
;                       0       130     137
;                  SAT
;                       1       130     137
;
;       FPA32           32 bit floating point add
;
;               Timing:            RND
;                               0       1
;
;                       0       251     265
;                  SAT
;                       1       251     271
;
;       FPS32           32 bit floating point subtract
;
;               Timing:            RND
;                               0       1
;
;                       0       253     267
;                  SAT
;                       1       253     273
;
;       FPM32           32 bit floating point multiply
;
;               Timing:            RND
;                               0       1
;
;                       0       574     588
;                  SAT
;                       1       574     591
;
;       FPD32           32 bit floating point divide
;
;               Timing:            RND
;                               0       1
;
;                       0       932     968
;                  SAT
;                       1       932     971
;
;
;**********************************************************************************************
;**********************************************************************************************
;
;       32 bit floating point representation
;
;       EXPONENT        8 bit biased exponent
;
;                       It is important to note that the use of biased exponents produces
;                       a unique representation of a floating point 0, given by
;                       EXP = HIGHBYTE = MIDBYTE = LOWBYTE = 0x00, with 0 being
;                       the only number with EXP = 0.
;
;       HIGHBYTE        8 bit most significant byte of fraction in sign-magnitude representation,
;                       with SIGN = MSB, implicit MSB = 1 and radix point to the right of MSB
;
;       MIDBYTE         8 bit middle significant byte of sign-magnitude fraction
;
;       LOWBYTE         8 bit least significant byte of sign-magnitude fraction
;
;       EXPONENT        HIGHBYTE        MIDBYTE         LOWBYTE
;
;       xxxxxxxx        S.xxxxxxx       xxxxxxxx        xxxxxxxx
;
;                        |
;                      RADIX
;                      POINT
;
;
;**********************************************************************************************
;**********************************************************************************************

;       Integer to float conversion

;       Input:  24 bit 2's complement integer right justified in AARGB0, AARGB1, AARGB2

;       Use:    CALL    FLO2432 or      CALL    FLO32

;       Output: 32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  FLOAT( AARG )

;       Max Timing:     14+90 = 104 clks                SAT = 0
;                       14+96 = 110 clks                SAT = 1

;       Min Timing:     6+28 = 34 clks                  AARG = 0
;                       6+18 = 24 clks

;       PM: 14+38 = 52                                  DM: 7

;----------------------------------------------------------------------------------------------

FLO2432
FLO32           MOVLW           D'23'+EXPBIAS		; initialize exponent and add bias
                MOVWF           EXP
                CLRF            SIGN
                BTFSS           AARGB0,MSB		; test sign
                GOTO            NRM3232
                COMF            AARGB2,F                ; if < 0, negate and set MSB in SIGN
                COMF            AARGB1,F
                COMF            AARGB0,F
                INCF            AARGB2,F
                BTFSC           _Z
                INCF            AARGB1,F
                BTFSC           _Z
                INCF            AARGB0,F
                BSF             SIGN,MSB

;**********************************************************************************************

;       Normalization routine

;       Input:  32 bit unnormalized floating point number in AEXP, AARGB0, AARGB1,
;               AARGB2, with sign in SIGN,MSB

;       Use:    CALL    NRM3232 or      CALL    NRM32

;       Output: 32 bit normalized floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  NORMALIZE( AARG )

;       Max Timing:     21+6+7*8+7 = 90 clks            SAT = 0
;                       21+6+7*8+1+12 = 96 clks SAT = 1

;       Min Timing:     22+6 = 28 clks                  AARG = 0
;                       5+9+4 = 18 clks

;       PM: 38                                          DM: 7

;----------------------------------------------------------------------------------------------

NRM3232

NRM32           CLRF            TEMP			; clear exponent decrement
                MOVF            AARGB0,W		; test if highbyte=0
                BTFSS           _Z
                GOTO            NORM3232
                MOVF            AARGB1,W		; if so, shift 8 bits by move
                MOVWF           AARGB0
                MOVF            AARGB2,W
                MOVWF           AARGB1
                CLRF            AARGB2
                BSF             TEMP,3                  ; increase decrement by 8

                MOVF            AARGB0,W		; test if highbyte=0
                BTFSS           _Z
                GOTO            NORM3232
                MOVF            AARGB1,W		; if so, shift 8 bits by move
                MOVWF           AARGB0
                CLRF            AARGB1
                BCF             TEMP,3                  ; increase decrement by 8
                BSF             TEMP,4
        
                MOVF            AARGB0,W		; if highbyte=0, result=0
                BTFSC           _Z
                GOTO            RES032

NORM3232        MOVF            TEMP,W
                SUBWF           EXP,F
                BTFSS           _Z
                BTFSS           _C
                GOTO            SETFUN32

                BCF             _C                      ; clear carry bit

NORM3232A       BTFSC           AARGB0,MSB		; if MSB=1, normalization done
                GOTO            FIXSIGN32
                RLF             AARGB2,F                ; otherwise, shift left and 
                RLF             AARGB1,F                ; decrement EXP
                RLF             AARGB0,F
                DECFSZ          EXP,F
                GOTO            NORM3232A

                GOTO            SETFUN32                ; underflow if EXP=0

FIXSIGN32       BTFSS           SIGN,MSB
                BCF             AARGB0,MSB              ; clear explicit MSB if positive
                RETLW           0

RES032          CLRF            AARGB0                  ; result equals zero
                CLRF            AARGB1
                CLRF            AARGB2
		CLRF		AARGB3
                CLRF            EXP
                RETLW           0

;**********************************************************************************************
;**********************************************************************************************

;       Integer to float conversion

;       Input:  32 bit 2's complement integer right justified in AARGB0, AARGB1, AARGB2,
;               AARGB3

;       Use:    CALL    FLO3232

;       Output: 32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  FLOAT( AARG )

;       Max Timing:     17+112 = 129 clks               RND = 0
;                       17+128 = 145 clks               RND = 1, SAT = 0
;                       17+135 = 152 clks               RND = 1, SAT = 1

;       Min Timing:     6+39 = 45 clks                  AARG = 0
;                       6+22 = 28 clks

;       PM: 17+66 = 83                                  DM: 8

;----------------------------------------------------------------------------------------------

FLO3232         MOVLW           D'31'+EXPBIAS		; initialize exponent and add bias
                MOVWF           EXP
                CLRF            SIGN
                BTFSS           AARGB0,MSB		; test sign
                GOTO            NRM4032
                COMF            AARGB3,F                ; if < 0, negate and set MSB in SIGN
                COMF            AARGB2,F
                COMF            AARGB1,F
                COMF            AARGB0,F
                INCF            AARGB3,F
                BTFSC           _Z
                INCF            AARGB2,F
                BTFSC           _Z
                INCF            AARGB1,F
                BTFSC           _Z
                INCF            AARGB0,F
                BSF             SIGN,MSB

;**********************************************************************************************

;       Normalization routine

;       Input:  40 bit unnormalized floating point number in AEXP, AARGB0, AARGB1,
;               AARGB2, AARGB3 with sign in SIGN,MSB

;       Use:    CALL    NRM4032

;       Output: 32 bit normalized floating point number in AEXP, AARGB0, AARGB1, AARGB2,
;               AARGB3

;       Result: AARG  <--  NORMALIZE( AARG )

;       Max Timing:     38+6*9+12+8 = 112 clks  RND = 0
;                       38+6*9+12+24 = 128 clks RND = 1, SAT = 0
;                       38+6*9+12+31 = 135 clks RND = 1, SAT = 1

;       Min Timing:     33+6 = 39 clks                  AARG = 0
;                       5+9+8 = 22 clks

;       PM: 66                                          DM: 8

;----------------------------------------------------------------------------------------------

NRM4032         CLRF            TEMP			; clear exponent decrement
                MOVF            AARGB0,W		; test if highbyte=0
                BTFSS           _Z
                GOTO            NORM4032
                MOVF            AARGB1,W		; if so, shift 8 bits by move
                MOVWF           AARGB0
                MOVF            AARGB2,W
                MOVWF           AARGB1
                MOVF            AARGB3,W
                MOVWF           AARGB2
                CLRF            AARGB3
                BSF             TEMP,3                  ; increase decrement by 8

                MOVF            AARGB0,W		; test if highbyte=0
                BTFSS           _Z
                GOTO            NORM4032
                MOVF            AARGB1,W		; if so, shift 8 bits by move
                MOVWF           AARGB0
                MOVF            AARGB2,W
                MOVWF           AARGB1
                CLRF            AARGB2
                BCF             TEMP,3                  ; increase decrement by 8
                BSF             TEMP,4
        
                MOVF            AARGB0,W		; test if highbyte=0
                BTFSS           _Z
                GOTO            NORM4032
                MOVF            AARGB1,W		; if so, shift 8 bits by move
                MOVWF           AARGB0
                CLRF            AARGB1
                BSF             TEMP,3                  ; increase decrement by 8
        
                MOVF            AARGB0,W		; if highbyte=0, result=0
                BTFSC           _Z
                GOTO            RES032

NORM4032        MOVF            TEMP,W
                SUBWF           EXP,F
                BTFSS           _Z
                BTFSS           _C
                GOTO            SETFUN32

                BCF             _C                      ; clear carry bit

NORM4032A       BTFSC           AARGB0,MSB		; if MSB=1, normalization done
                GOTO            NRMRND4032
                RLF             AARGB3,F                ; otherwise, shift left and 
                RLF             AARGB2,F                ; decrement EXP
                RLF             AARGB1,F
                RLF             AARGB0,F
                DECFSZ          EXP,F
                GOTO            NORM4032A

                GOTO            SETFUN32                ; underflow if EXP=0

NRMRND4032      BTFSC           FPFLAGS,RND
                BTFSS           AARGB2,LSB
                GOTO            FIXSIGN32
		BTFSS		AARGB3,MSB		; round if next bit is set
                GOTO            FIXSIGN32
		INCF		AARGB2,F
                BTFSC           _Z
                INCF            AARGB1,F
                BTFSC           _Z
                INCF            AARGB0,F

                BTFSS           _Z                      ; has rounding caused carryout?
                GOTO            FIXSIGN32
                RRF             AARGB0,F                ; if so, right shift
                RRF             AARGB1,F
                RRF             AARGB2,F
                INCF            EXP,F
                BTFSC           _Z                      ; check for overflow
                GOTO            SETFOV32
                GOTO            FIXSIGN32

;**********************************************************************************************
;**********************************************************************************************

;       Float to integer conversion

;       Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Use:    CALL    INT3224         or      CALL    INT32

;       Output: 24 bit 2's complement integer right justified in AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  INT( AARG )

;       Max Timing:     40+6*7+6+16 = 104 clks		RND = 0
;                       40+6*7+6+24 = 112 clks		RND = 1, SAT = 0
;                       40+6*7+6+26 = 114 clks  	RND = 1, SAT = 1

;       Min Timing:     4 clks

;       PM: 82                                          DM: 6

;----------------------------------------------------------------------------------------------

INT3224
INT32
		MOVF		EXP,W			; test for zero argument
		BTFSC		_Z
		RETLW		0x00

		MOVF            AARGB0,W		; save sign in SIGN
                MOVWF           SIGN
                BSF             AARGB0,MSB		; make MSB explicit

                MOVLW           EXPBIAS+D'23'		; remove bias from EXP
                SUBWF           EXP,F
                BTFSS           EXP,MSB
                GOTO            SETIOV3224
		COMF		EXP,F
		INCF		EXP,F

                MOVLW           8                       ; do byte shift if EXP >= 8
                SUBWF           EXP,W
                BTFSS           _C
                GOTO            TSHIFT3224
                MOVWF           EXP
                RLF             AARGB2,F                ; rotate next bit for rounding
                MOVF            AARGB1,W
                MOVWF           AARGB2
                MOVF            AARGB0,W
                MOVWF           AARGB1
                CLRF            AARGB0

                MOVLW           8                       ; do another byte shift if EXP >= 8
                SUBWF           EXP,W
                BTFSS           _C
                GOTO            TSHIFT3224
                MOVWF           EXP
                RLF             AARGB2,F		; rotate next bit for rounding
                MOVF            AARGB1,W
                MOVWF           AARGB2
                CLRF            AARGB1

                MOVLW           8                       ; do another byte shift if EXP >= 8
                SUBWF           EXP,W
                BTFSS           _C
                GOTO            TSHIFT3224
                MOVWF           EXP
                RLF             AARGB2,F                ; rotate next bit for rounding
                CLRF            AARGB2
		MOVF		EXP,W
		BTFSS		_Z
		BCF		_C
		GOTO		SHIFT3224OK

TSHIFT3224      MOVF            EXP,W                   ; shift completed if EXP = 0
                BTFSC           _Z
                GOTO            SHIFT3224OK

SHIFT3224       BCF             _C
                RRF             AARGB0,F                ; right shift by EXP
                RRF             AARGB1,F
                RRF             AARGB2,F
                DECFSZ          EXP,F
                GOTO            SHIFT3224

SHIFT3224OK     BTFSC           FPFLAGS,RND
                BTFSS           AARGB2,LSB
                GOTO            INT3224OK
                BTFSS           _C
                GOTO            INT3224OK
                INCF            AARGB2,F
                BTFSC           _Z
                INCF            AARGB1,F
                BTFSC           _Z
                INCF            AARGB0,F
                BTFSC           AARGB0,MSB		; test for overflow
                GOTO            SETIOV3224

INT3224OK       BTFSS           SIGN,MSB                ; if sign bit set, negate               
                RETLW           0
                COMF            AARGB0,F
                COMF            AARGB1,F
                COMF            AARGB2,F
                INCF            AARGB2,F
                BTFSC           _Z
                INCF            AARGB1,F
                BTFSC           _Z
                INCF            AARGB0,F
                RETLW           0

IRES03224	CLRF            AARGB0			; integer result equals zero
                CLRF            AARGB1
                CLRF            AARGB2
                RETLW           0

SETIOV3224	BSF             FPFLAGS,IOV             ; set integer overflow flag
                BTFSS           FPFLAGS,SAT             ; test for saturation
                RETLW           0xFF                    ; return error code in WREG

                CLRF            AARGB0			; saturate to largest two's
                BTFSS           SIGN,MSB                ; complement 24 bit integer
                MOVLW           0xFF
                MOVWF           AARGB0			; SIGN = 0, 0x 7F FF FF
                MOVWF           AARGB1			; SIGN = 1, 0x 80 00 00
                MOVWF           AARGB2
                RLF             SIGN,F
                RRF             AARGB0,F
                RETLW           0xFF                    ; return error code in WREG

;**********************************************************************************************
;**********************************************************************************************

;       Float to integer conversion

;       Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Use:    CALL    INT3232

;       Output: 32 bit 2's complement integer right justified in AARGB0, AARGB1, AARGB2,
;               AARGB3

;       Result: AARG  <--  INT( AARG )

;       Max Timing:     54+6*8+7+21 = 130 clks          RND = 0
;                       54+6*8+7+29 = 137 clks          RND = 1, SAT = 0
;                       54+6*8+7+29 = 137 clks          RND = 1, SAT = 1

;       Min Timing:     5 clks

;       PM: 102                                                 DM: 7

;----------------------------------------------------------------------------------------------

INT3232
                CLRF            AARGB3
		MOVF		EXP,W			; test for zero argument
		BTFSC		_Z
		RETLW		0x00

		MOVF            AARGB0,W		; save sign in SIGN
                MOVWF           SIGN
                BSF             AARGB0,MSB		; make MSB explicit

                MOVLW           EXPBIAS+D'31'		; remove bias from EXP
                SUBWF           EXP,F
                BTFSS           EXP,MSB
                GOTO            SETIOV32
		COMF		EXP,F
		INCF		EXP,F        

                MOVLW           8                       ; do byte shift if EXP >= 8
                SUBWF           EXP,W
                BTFSS           _C
                GOTO            TSHIFT3232
                MOVWF           EXP
                RLF             AARGB3,F                ; rotate next bit for rounding
                MOVF            AARGB2,W
                MOVWF           AARGB3
                MOVF            AARGB1,W
                MOVWF           AARGB2
                MOVF            AARGB0,W
                MOVWF           AARGB1
                CLRF            AARGB0

                MOVLW           8                       ; do another byte shift if EXP >= 8
                SUBWF           EXP,W
                BTFSS           _C
                GOTO            TSHIFT3232
                MOVWF           EXP
                RLF             AARGB3,F                ; rotate next bit for rounding
                MOVF            AARGB2,W
                MOVWF           AARGB3
                MOVF            AARGB1,W
                MOVWF           AARGB2
                CLRF            AARGB1

                MOVLW           8                       ; do another byte shift if EXP >= 8
                SUBWF           EXP,W
                BTFSS           _C
                GOTO            TSHIFT3232
                MOVWF           EXP
                RLF             AARGB3,F                ; rotate next bit for rounding
                MOVF            AARGB2,W
                MOVWF           AARGB3
                CLRF            AARGB2

                MOVLW           8                       ; do another byte shift if EXP >= 8
                SUBWF           EXP,W
                BTFSS           _C
                GOTO            TSHIFT3232
                MOVWF           EXP
                RLF             AARGB3,F                ; rotate next bit for rounding
                CLRF            AARGB3
		MOVF		EXP,W
		BTFSS		_Z
		BCF		_C
		GOTO		SHIFT3232OK

TSHIFT3232      MOVF            EXP,W                   ; shift completed if EXP = 0
                BTFSC           _Z
                GOTO            SHIFT3232OK

SHIFT3232       BCF             _C
                RRF             AARGB0,F                ; right shift by EXP
                RRF             AARGB1,F
                RRF             AARGB2,F
                RRF             AARGB3,F
                DECFSZ          EXP,F
                GOTO            SHIFT3232

SHIFT3232OK     BTFSC           FPFLAGS,RND
                BTFSS           AARGB3,LSB
                GOTO            INT3232OK
                BTFSS           _C
                GOTO            INT3232OK
                INCF            AARGB3,F
                BTFSC           _Z
                INCF            AARGB2,F
                BTFSC           _Z
                INCF            AARGB1,F
                BTFSC           _Z
                INCF            AARGB0,F
                BTFSC           AARGB0,MSB		; test for overflow
                GOTO            SETIOV3224

INT3232OK       BTFSS           SIGN,MSB                ; if sign bit set, negate               
                RETLW           0
                COMF            AARGB0,F
                COMF            AARGB1,F
                COMF            AARGB2,F
                COMF            AARGB3,F
                INCF            AARGB3,F
                BTFSC           _Z
                INCF            AARGB2,F
                BTFSC           _Z
                INCF            AARGB1,F
                BTFSC           _Z
                INCF            AARGB0,F
                RETLW           0

IRES032         CLRF            AARGB0			; integer result equals zero
                CLRF            AARGB1
                CLRF            AARGB2
                CLRF            AARGB3
                RETLW           0

SETIOV32        BSF             FPFLAGS,IOV             ; set integer overflow flag
                BTFSS           FPFLAGS,SAT             ; test for saturation
                RETLW           0xFF                    ; return error code in WREG

                CLRF            AARGB0			; saturate to largest two's
                BTFSS           SIGN,MSB                ; complement 32 bit integer
                MOVLW           0xFF
                MOVWF           AARGB0			; SIGN = 0, 0x 7F FF FF FF
                MOVWF           AARGB1			; SIGN = 1, 0x 80 00 00 00
                MOVWF           AARGB2
                MOVWF           AARGB3
                RLF             SIGN,F
                RRF             AARGB0,F
                RETLW           0xFF                    ; return error code in WREG

;**********************************************************************************************
;**********************************************************************************************

;       Floating Point Multiply

;       Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2
;               32 bit floating point number in BEXP, BARGB0, BARGB1, BARGB2

;       Use:    CALL    FPM32

;       Output: 32 bit floating point product in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  AARG * BARG

;       Max Timing:     26+23*22+21+21 = 574 clks       RND = 0
;                       26+23*22+21+35 = 588 clks       RND = 1, SAT = 0
;                       26+23*22+21+38 = 591 clks       RND = 1, SAT = 1

;       Min Timing:     6+6 = 12 clks                   AARG * BARG = 0
;                       24+23*11+21+17 = 315 clks

;       PM: 94                                          DM: 14

;----------------------------------------------------------------------------------------------

FPM32           MOVF            AEXP,W                  ; test for zero arguments
                BTFSS           _Z
                MOVF            BEXP,W
                BTFSC           _Z
                GOTO            RES032

M32BNE0         MOVF            AARGB0,W
                XORWF           BARGB0,W
                MOVWF           SIGN                    ; save sign in SIGN

                MOVF            BEXP,W
                ADDWF           EXP,F
                MOVLW           EXPBIAS-1
                BTFSS           _C
                GOTO            MTUN32

                SUBWF           EXP,F
                BTFSC           _C
                GOTO            SETFOV32                ; set multiply overflow flag
                GOTO            MOK32

MTUN32          SUBWF           EXP,F
                BTFSS           _C
                GOTO            SETFUN32

MOK32		MOVF		AARGB0,W
		MOVWF		AARGB3
		MOVF		AARGB1,W
		MOVWF		AARGB4
		MOVF		AARGB2,W
		MOVWF		AARGB5
		BSF             AARGB3,MSB              ; make argument MSB's explicit
                BSF             BARGB0,MSB
                BCF             _C
                CLRF            AARGB0			; clear initial partial product
                CLRF            AARGB1
                CLRF            AARGB2
                MOVLW           D'24'
                MOVWF           TEMP                    ; initialize counter

MLOOP32         BTFSS           AARGB5,LSB              ; test next bit
                GOTO            MNOADD32

MADD32          MOVF            BARGB2,W
                ADDWF           AARGB2,F
                MOVF            BARGB1,W
                BTFSC           _C
                INCFSZ          BARGB1,W
                ADDWF           AARGB1,F

                MOVF            BARGB0,W
                BTFSC           _C
                INCFSZ          BARGB0,W
                ADDWF           AARGB0,F

MNOADD32        RRF             AARGB0,F
                RRF             AARGB1,F
                RRF             AARGB2,F
                RRF             AARGB3,F
                RRF             AARGB4,F
                RRF             AARGB5,F
                BCF             _C
                DECFSZ          TEMP,F
                GOTO            MLOOP32

                BTFSC           AARGB0,MSB               ; check for postnormalization
                GOTO            MROUND32
                RLF             AARGB3,F
                RLF             AARGB2,F
                RLF             AARGB1,F
                RLF             AARGB0,F
                DECF            EXP,F

MROUND32        BTFSC           FPFLAGS,RND
                BTFSS           AARGB2,LSB
                GOTO            MUL32OK
		BTFSS		AARGB3,MSB
                GOTO            MUL32OK
		INCF		AARGB2,F
                BTFSC           _Z
                INCF            AARGB1,F
                BTFSC           _Z
                INCF            AARGB0,F

                BTFSS           _Z                      ; has rounding caused carryout?
                GOTO            MUL32OK
                RRF             AARGB0,F                ; if so, right shift
                RRF             AARGB1,F
                RRF             AARGB2,F
                INCF            EXP,F
                BTFSC           _Z                      ; check for overflow
                GOTO            SETFOV32

MUL32OK         BTFSS           SIGN,MSB
                BCF             AARGB0,MSB		; clear explicit MSB if positive

                RETLW           0  

SETFOV32        BSF             FPFLAGS,FOV             ; set floating point underflag
                BTFSS           FPFLAGS,SAT             ; test for saturation
                RETLW           0xFF                    ; return error code in WREG

                MOVLW           0xFF
                MOVWF           AEXP                    ; saturate to largest floating
                MOVWF           AARGB0                  ; point number = 0x FF 7F FF FF
                MOVWF           AARGB1                  ; modulo the appropriate sign bit
                MOVWF           AARGB2
                RLF             SIGN,F
                RRF             AARGB0,F
                RETLW           0xFF                    ; return error code in WREG

;**********************************************************************************************
;**********************************************************************************************

;       Floating Point Divide

;       Input:  32 bit floating point dividend in AEXP, AARGB0, AARGB1, AARGB2
;               32 bit floating point divisor in BEXP, BARGB0, BARGB1, BARGB2

;       Use:    CALL    FPD32

;       Output: 32 bit floating point quotient in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  AARG / BARG

;       Max Timing:     43+12+23*36+35+14 = 932 clks            RND = 0
;                       43+12+23*36+35+50 = 968 clks            RND = 1, SAT = 0
;                       43+12+23*36+35+53 = 971 clks            RND = 1, SAT = 1

;       Min Timing:     7+6 = 13 clks

;       PM: 155                                                 DM: 14

;----------------------------------------------------------------------------------------------

FPD32           MOVF            BEXP,W                  ; test for divide by zero
                BTFSC           _Z
                GOTO            SETFDZ32

                MOVF            AEXP,W
                BTFSC           _Z
                GOTO            RES032

D32BNE0         MOVF            AARGB0,W
                XORWF           BARGB0,W
                MOVWF           SIGN                    ; save sign in SIGN
                BSF             AARGB0,MSB              ; make argument MSB's explicit
                BSF             BARGB0,MSB

TALIGN32        CLRF            TEMP                    ; clear align increment
                MOVF            AARGB0,W
                MOVWF           AARGB3			; test for alignment
                MOVF            AARGB1,W
                MOVWF           AARGB4
                MOVF            AARGB2,W
                MOVWF           AARGB5

                MOVF            BARGB2,W
                SUBWF           AARGB5,F
                MOVF            BARGB1,W
                BTFSS           _C
                INCFSZ          BARGB1,W

TS1ALIGN32      SUBWF           AARGB4,F
                MOVF            BARGB0,W
                BTFSS           _C
                INCFSZ          BARGB0,W

TS2ALIGN32      SUBWF           AARGB3,F

                CLRF            AARGB3
                CLRF            AARGB4
                CLRF            AARGB5

                BTFSS           _C
                GOTO            DALIGN32OK

                BCF             _C                      ; align if necessary
                RRF             AARGB0,F
                RRF             AARGB1,F
                RRF             AARGB2,F
                RRF             AARGB3,F
                MOVLW           0x01
                MOVWF           TEMP                    ; save align increment          

DALIGN32OK      MOVF            BEXP,W                  ; compare AEXP and BEXP
                SUBWF           EXP,F
                BTFSS           _C
                GOTO            ALTB32
        
AGEB32          MOVLW           EXPBIAS-1
                ADDWF           TEMP,W
                ADDWF           EXP,F
                BTFSC           _C
                GOTO            SETFOV32
                GOTO            DARGOK32                ; set overflow flag

ALTB32          MOVLW           EXPBIAS-1
                ADDWF           TEMP,W
                ADDWF           EXP,F
                BTFSS           _C
                GOTO            SETFUN32                ; set underflow flag

DARGOK32        MOVLW           D'24'			; initialize counter
                MOVWF           TEMPB1

DLOOP32         RLF             AARGB5,F                ; left shift
                RLF             AARGB4,F
                RLF             AARGB3,F
                RLF             AARGB2,F
                RLF             AARGB1,F
                RLF             AARGB0,F
                RLF             TEMP,F

                MOVF            BARGB2,W                ; subtract
                SUBWF           AARGB2,F
                MOVF            BARGB1,W
                BTFSS           _C
                INCFSZ          BARGB1,W
DS132           SUBWF           AARGB1,F

                MOVF            BARGB0,W
                BTFSS           _C
                INCFSZ          BARGB0,W
DS232           SUBWF           AARGB0,F

                RLF             BARGB0,W
                IORWF           TEMP,F
                
                BTFSS           TEMP,LSB                ; test for restore
                GOTO            DREST32

                BSF             AARGB5,LSB
                GOTO            DOK32

DREST32         MOVF            BARGB2,W                ; restore if necessary
                ADDWF           AARGB2,F
                MOVF            BARGB1,W
                BTFSC           _C
                INCFSZ          BARGB1,W
DAREST32        ADDWF           AARGB1,F

                MOVF            BARGB0,W
                BTFSC           _C
                INCF            BARGB0,W
                ADDWF           AARGB0,F

                BCF             AARGB5,LSB

DOK32           DECFSZ          TEMPB1,F
                GOTO            DLOOP32

DROUND32        BTFSC           FPFLAGS,RND
                BTFSS           AARGB5,LSB
                GOTO            DIV32OK
                BCF             _C
                RLF             AARGB2,F               ; compute next significant bit
                RLF             AARGB1,F               ; for rounding
                RLF             AARGB0,F
                RLF             TEMP,F

                MOVF            BARGB2,W               ; subtract
                SUBWF           AARGB2,F
                MOVF            BARGB1,W
                BTFSS           _C
                INCFSZ          BARGB1,W
		SUBWF           AARGB1,F

                MOVF            BARGB0,W
                BTFSS           _C
                INCFSZ          BARGB0,W
		SUBWF           AARGB0,F

                RLF             BARGB0,W
                IORWF           TEMP,W
                ANDLW           0x01            

                ADDWF           AARGB5,F
                BTFSC           _C
                INCF            AARGB4,F
                BTFSC           _Z
                INCF            AARGB3,F

                BTFSS           _Z                      ; test if rounding caused carryout
                GOTO            DIV32OK
                RRF             AARGB3,F
                RRF             AARGB4,F
                RRF             AARGB5,F
                INCF            EXP,F
                BTFSC           _Z                      ; test for overflow
                GOTO            SETFOV32


DIV32OK         BTFSS           SIGN,MSB
                BCF             AARGB3,MSB		; clear explicit MSB if positive

                MOVF            AARGB3,W
                MOVWF           AARGB0                  ; move result to AARG
                MOVF            AARGB4,W
                MOVWF           AARGB1
                MOVF            AARGB5,W
                MOVWF           AARGB2

                RETLW           0

SETFUN32        BSF             FPFLAGS,FUN             ; set floating point underflag
                BTFSS           FPFLAGS,SAT             ; test for saturation
                RETLW           0xFF                    ; return error code in WREG

                MOVLW           0x01                    ; saturate to smallest floating
                MOVWF           AEXP                    ; point number = 0x 01 00 00 00
                CLRF            AARGB0                  ; modulo the appropriate sign bit
                CLRF            AARGB1
                CLRF            AARGB2
                RLF             SIGN,F
                RRF             AARGB0,F
                RETLW           0xFF                    ; return error code in WREG

SETFDZ32        BSF             FPFLAGS,FDZ             ; set divide by zero flag
                RETLW           0xFF

;**********************************************************************************************
;**********************************************************************************************

;       Floating Point Subtract

;       Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2
;               32 bit floating point number in BEXP, BARGB0, BARGB1, BARGB2

;       Use:    CALL FPS32

;       Output: 32 bit floating point sum in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  AARG - BARG

;       Max Timing:     2+251 = 253 clks                RND = 0
;                       2+265 = 267 clks                RND = 1, SAT = 0
;                       2+271 = 273 clks                RND = 1, SAT = 1

;       Min Timing:     2+12 = 14 clks

;       PM: 2+146 = 148                         DM: 14

;----------------------------------------------------------------------------------------------

FPS32           MOVLW           0x80
                XORWF           BARGB0,F

;**********************************************************************************************

;       Floating Point Add

;       Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2
;               32 bit floating point number in BEXP, BARGB0, BARGB1, BARGB2

;       Use:    CALL FPA32

;       Output: 32 bit floating point sum in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  AARG - BARG

;       Max Timing:     31+41+6*7+6+41+90 = 251 clks            RND = 0
;                       31+41+6*7+6+55+90 = 265 clks            RND = 1, SAT = 0
;                       31+41+6*7+6+55+96 = 271 clks            RND = 1, SAT = 1

;       Min Timing:     8+4 = 12 clks

;       PM: 146                                                 DM: 14

;----------------------------------------------------------------------------------------------

FPA32           MOVF            AARGB0,W                ; exclusive or of signs in TEMP
                XORWF           BARGB0,W
                MOVWF           TEMP

		CLRF		AARGB3			; clear extended byte
		CLRF		BARGB3

                MOVF            AEXP,W                  ; use AARG if AEXP >= BEXP
                SUBWF           BEXP,W
                BTFSS           _C
                GOTO            USEA32

                MOVF            BEXP,W                  ; use BARG if AEXP < BEXP
                MOVWF           AARGB5			; therefore, swap AARG and BARG
                MOVF            AEXP,W
                MOVWF           BEXP
                MOVF            AARGB5,W
                MOVWF           AEXP

                MOVF            BARGB0,W
                MOVWF           AARGB5
                MOVF            AARGB0,W
                MOVWF           BARGB0
                MOVF            AARGB5,W
                MOVWF           AARGB0

                MOVF            BARGB1,W
                MOVWF           AARGB5
                MOVF            AARGB1,W
                MOVWF           BARGB1
                MOVF            AARGB5,W
                MOVWF           AARGB1

                MOVF            BARGB2,W
                MOVWF           AARGB5
                MOVF            AARGB2,W
                MOVWF           BARGB2
                MOVF            AARGB5,W
                MOVWF           AARGB2

USEA32          MOVF            BEXP,W                  ; return AARG if BARG = 0
                BTFSC           _Z
                RETLW           0x00

                MOVF            AARGB0,W
                MOVWF           SIGN                    ; save sign in SIGN
                BSF             AARGB0,MSB              ; make MSB's explicit
                BSF             BARGB0,MSB

                MOVF            BEXP,W                  ; compute shift count in BEXP
                SUBWF           AEXP,W
                MOVWF           BEXP
                BTFSC           _Z
                GOTO            ALIGNED32

                MOVLW           8
                SUBWF           BEXP,W
                BTFSS           _C                      ; if BEXP >= 8, do byte shift
                GOTO            ALIGNB32
                MOVWF           BEXP
                MOVF            BARGB2,W		; keep for postnormalization
		MOVWF		BARGB3
                MOVF            BARGB1,W
		MOVWF		BARGB2
                MOVF            BARGB0,W
		MOVWF		BARGB1
                CLRF            BARGB0

                MOVLW           8
                SUBWF           BEXP,W
                BTFSS           _C                      ; if BEXP >= 8, do byte shift
                GOTO            ALIGNB32
                MOVWF           BEXP
                MOVF            BARGB2,W		; keep for postnormalization
		MOVWF		BARGB3
                MOVF            BARGB1,W
		MOVWF		BARGB2
                CLRF            BARGB1

                MOVLW           8
                SUBWF           BEXP,W
                BTFSS           _C                      ; if BEXP >= 8, BARG = 0 relative to AARG
                GOTO            ALIGNB32
                MOVF            SIGN,W
                MOVWF           AARGB0
                RETLW           0x00

ALIGNB32        MOVF            BEXP,W                  ; already aligned if BEXP = 0
                BTFSC           _Z
                GOTO            ALIGNED32

ALOOPB32        BCF             _C                      ; right shift by BEXP
                RRF             BARGB0,F
                RRF             BARGB1,F
		RRF		BARGB2,F
		RRF		BARGB3,F
                DECFSZ          BEXP,F
                GOTO            ALOOPB32

ALIGNED32       BTFSS           TEMP,MSB                ; negate if signs opposite
                GOTO            AOK32

		COMF		BARGB3,F
		COMF		BARGB2,F
                COMF            BARGB1,F
                COMF            BARGB0,F
                INCF            BARGB3,F
                BTFSC           _Z
                INCF            BARGB2,F
		BTFSC		_Z
		INCF		BARGB1,F
		BTFSC		_Z
		INCF		BARGB0,F

AOK32
                MOVF   		BARGB3,W
                ADDWF   	AARGB3,F
                MOVF            BARGB2,W
                BTFSC           _C
                INCFSZ          BARGB2,W
                ADDWF           AARGB2,F
                MOVF            BARGB1,W
                BTFSC           _C
                INCFSZ          BARGB1,W
                ADDWF           AARGB1,F
                MOVF            BARGB0,W
                BTFSC           _C
                INCFSZ          BARGB0,W
                ADDWF           AARGB0,F

                BTFSC           TEMP,MSB
                GOTO            ACOMP32
                BTFSS           _C
                GOTO            NRMRND4032

                RRF             AARGB0,F               ; shift right and increment EXP
                RRF             AARGB1,F
                RRF             AARGB2,F
		RRF		AARGB3,F
                INCFSZ          AEXP,F
                GOTO            NRMRND4032
                GOTO            SETFOV32

ACOMP32         BTFSC           _C
                GOTO            NRM4032			; normalize and fix sign

		COMF		AARGB3,F
                COMF            AARGB2,F		; negate, toggle sign bit and
                COMF            AARGB1,F		; then normalize
                COMF            AARGB0,F
                INCF            AARGB3,F
                BTFSC           _Z
                INCF            AARGB2,F
                BTFSC           _Z
                INCF            AARGB1,F
                BTFSC           _Z
                INCF            AARGB0,F

                MOVLW           0x80
                XORWF           SIGN,F
                GOTO            NRM32

;	RCS Header $Id: fxd68.a16 2.3 1996/10/16 14:23:57 F.J.Testa Exp $

;	$Revision: 2.3 $

;       16/8 PIC16 FIXED POINT DIVIDE ROUTINES
;
;       Input:  fixed point arguments in AARG and BARG
;
;       Output: quotient AARG/BARG followed by remainder in REM
;
;       All timings are worst case cycle counts
;
;       It is useful to note that the additional unsigned routines requiring a non-power of two
;       argument can be called in a signed divide application where it is known that the
;       respective argument is nonnegative, thereby offering some improvement in
;       performance.
;
;         Routine            Clocks     Function
;
;       FXD1608S    203 16 bit/8 bit -> 16.08 signed fixed point divide
;
;       FXD1608U    294 16 bit/8 bit -> 16.08 unsigned fixed point divide
;
;       FXD1607U    174 16 bit/7 bit -> 16.07 unsigned fixed point divide
;
;       FXD1507U    166 15 bit/7 bit -> 15.07 unsigned fixed point divide
;
;       The above timings are based on the looped macros. If space permits,
;       approximately 41-50 clocks can be saved by using the unrolled macros.
;

;**********************************************************************************************
;**********************************************************************************************

;       16/08 BIT Division Macros


SDIV1608L       macro

;       Max Timing:     3+5+2+5*11+10+10+6*11+10+2 = 163 clks

;       Min Timing:     3+5+2+5*11+10+10+6*11+10+2 = 163 clks

;       PM: 42                                  DM: 5

                MOVF            BARGB0,W
                SUBWF           REMB0, F
                RLF             AARGB0, F

                RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W
                ADDWF           REMB0, F
                RLF             AARGB0, F

                MOVLW           6
                MOVWF           LOOPCNT

LOOPS1608A      RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB0, F

                DECFSZ          LOOPCNT, F
                GOTO            LOOPS1608A

                RLF             AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                MOVLW           7
                MOVWF           LOOPCNT

LOOPS1608B      RLF             AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB1,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                DECFSZ          LOOPCNT, F
                GOTO            LOOPS1608B

                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F

                endm

UDIV1608L  macro

;       Max Timing: 2+7*12+11+3+7*24+23 = 291 clks

;       Min Timing: 2+7*11+10+3+7*17+16 = 227 clks

;       PM: 39                                  DM: 7

                MOVLW           8
                MOVWF           LOOPCNT

LOOPU1608A      RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W
                SUBWF           REMB0, F

                BTFSC           _C
                GOTO            UOK68A          
                ADDWF           REMB0, F
                BCF             _C
UOK68A          RLF             AARGB0, F

                DECFSZ          LOOPCNT, F
                GOTO            LOOPU1608A

                CLRF            TEMP

                MOVLW           8
                MOVWF           LOOPCNT

LOOPU1608B      RLF             AARGB1,W
                RLF             REMB0, F
                RLF             TEMP, F
                MOVF            BARGB0,W
                SUBWF           REMB0, F
                CLRF            AARGB5
                CLRW
                BTFSS           _C
                INCFSZ          AARGB5,W
                SUBWF           TEMP, F

                BTFSC           _C
                GOTO            UOK68B          
                MOVF            BARGB0,W
                ADDWF           REMB0, F
                CLRF            AARGB5
                CLRW
                BTFSC           _C
                INCFSZ          AARGB5,W
                ADDWF           TEMP, F

                BCF             _C
UOK68B          RLF             AARGB1, F

                DECFSZ          LOOPCNT, F
                GOTO            LOOPU1608B

                endm

UDIV1607L       macro

;       Max Timing:     7+6*11+10+10+6*11+10+2 = 171 clks

;       Min Timing:     7+6*11+10+10+6*11+10+2 = 171 clks

;       PM: 39                                  DM: 5

                RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W
                SUBWF           REMB0, F
                RLF             AARGB0, F

                MOVLW           7
                MOVWF           LOOPCNT

LOOPU1607A      RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB0, F

                DECFSZ          LOOPCNT, F
                GOTO            LOOPU1607A

                RLF             AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                MOVLW           7
                MOVWF           LOOPCNT

LOOPU1607B 	RLF		AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB1,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                DECFSZ          LOOPCNT, F
                GOTO            LOOPU1607B

                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F

                endm

UDIV1507L       macro

;       Max Timing:     3+5+2+5*11+10+10+6*11+10+2 = 163 clks

;       Min Timing:     3+5+2+5*11+10+10+6*11+10+2 = 163 clks

;       PM: 42                                  DM: 5

                MOVF            BARGB0,W
                SUBWF           REMB0, F
                RLF             AARGB0, F

                RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W
                ADDWF           REMB0, F
                RLF             AARGB0, F

                MOVLW           6
                MOVWF           LOOPCNT

LOOPU1507A      RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB0, F

                DECFSZ          LOOPCNT, F
                GOTO            LOOPU1507A

                RLF             AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                MOVLW           7
                MOVWF           LOOPCNT

LOOPU1507B      RLF             AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB1,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                DECFSZ          LOOPCNT, F
                GOTO            LOOPU1507B

                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F

                endm

SDIV1608        macro

;       Max Timing:     3+5+14*8+2 = 122 clks

;       Min Timing:     3+5+14*8+2 = 122 clks

;       PM: 122                                 DM: 4

                MOVF            BARGB0,W
                SUBWF           REMB0, F
                RLF             AARGB0, F

                RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W
                ADDWF           REMB0, F
                RLF             AARGB0, F

                variable i = 2

                while i < 8

                RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB0, F

                variable i =i+1

                endw

                RLF             AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                variable i = 9

                while i < 16

                RLF             AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB1,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                variable i =i+1

                endw

                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F

                endm

UDIV1608  macro

;       restore = 9/21 clks,  nonrestore = 8/14 clks

;       Max Timing: 8*9+1+8*21 = 241 clks

;       Min Timing: 8*8+1+8*14 = 177 clks

;       PM: 241                                 DM: 6

                variable i = 0

                while i < 8

                RLF             AARGB0,W
                RLF              REMB0, F
                MOVF            BARGB0,W
                SUBWF           REMB0, F

                BTFSC           _C
                GOTO            UOK68#v(i)              
                ADDWF           REMB0, F
                BCF             _C
UOK68#v(i)      RLF     AARGB0, F

                variable i =i+1

                endw

                CLRF            TEMP

                variable i = 8

                while i < 16

                RLF             AARGB1,W
                RLF             REMB0, F
                RLF             TEMP, F
                MOVF            BARGB0,W
                SUBWF           REMB0, F
                CLRF            AARGB5
                CLRW
                BTFSS           _C
                INCFSZ          AARGB5,W
                SUBWF           TEMP, F

                BTFSC           _C
                GOTO            UOK68#v(i)              
                MOVF            BARGB0,W
                ADDWF           REMB0, F
                CLRF            AARGB5
                CLRW
                BTFSC           _C
                INCFSZ          AARGB5,W
                ADDWF           TEMP, F

                BCF             _C
UOK68#v(i)      RLF             AARGB1, F

                variable i =i+1

                endw

                endm

UDIV1607        macro

;       Max Timing:     5+15*8+2 = 127 clks

;       Min Timing:     5+15*8+2 = 127 clks

;       PM: 127                                 DM: 4

                RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W
                SUBWF           REMB0, F
                RLF             AARGB0, F

                variable i = 1

                while i < 8

                RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB0, F

                variable i =i+1

                endw

                RLF             AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                variable i = 9

                while i < 16

                RLF             AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB1,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                variable i =i+1

                endw

                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F

                endm

UDIV1507        macro

;       Max Timing:     3+5+14*8+2 = 122 clks

;       Min Timing:     3+5+14*8+2 = 122 clks

;       PM: 122                                 DM: 4

                MOVF            BARGB0,W
                SUBWF           REMB0, F
                RLF             AARGB0, F

                RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W
                ADDWF           REMB0, F
                RLF             AARGB0, F

                variable i = 2

                while i < 8

                RLF             AARGB0,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB0, F

                variable i =i+1

                endw

                RLF             AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB0,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB0,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                variable i = 9

                while i < 16

                RLF             AARGB1,W
                RLF             REMB0, F
                MOVF            BARGB0,W

                BTFSC           AARGB1,LSB
                SUBWF           REMB0, F
                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F
                RLF             AARGB1, F

                variable i =i+1

                endw

                BTFSS           AARGB1,LSB
                ADDWF           REMB0, F

                endm
                
;**********************************************************************************************
;**********************************************************************************************
        
;       16/8 Bit Signed Fixed Point Divide 16/8 -> 16.08

;       Input:  16 bit signed fixed point dividend in AARGB0, AARGB1
;               8 bit signed fixed point divisor in BARGB0

;       Use:    CALL    FXD1608S

;       Output: 16 bit signed fixed point quotient in AARGB0, AARGB1
;               8 bit signed fixed point remainder in REMB0

;       Result: AARG, REM  <--  AARG / BARG

;       Max Timing:     23+163+5 = 191 clks             A > 0, B > 0
;                       24+163+13 = 200 clks            A > 0, B < 0
;                       27+163+13 = 203 clks            A < 0, B > 0
;                       28+163+5 = 196 clks             A < 0, B < 0
;			7 clks				A = 0

;       Min Timing:     23+163+5 = 191 clks             A > 0, B > 0
;                       24+163+13 = 200 clks            A > 0, B < 0
;                       27+163+13 = 203 clks            A < 0, B > 0
;                       28+163+5 = 196 clks             A < 0, B < 0

;       PM: 28+42+12+34 = 116               DM: 8

FXD1608S	CLRF		SIGN
                CLRF            REMB0			; clear partial remainder
		MOVF		AARGB0,W
		IORWF		AARGB1,W
		BTFSC		_Z
		RETLW		0x00

		MOVF		AARGB0,W
		XORWF		BARGB0,W
		MOVWF		TEMP
		BTFSC		TEMP,MSB
		COMF		SIGN,F

		CLRF		TEMPB3			; clear exception flag

                BTFSS           BARGB0,MSB        ; if MSB set, negate BARG
                GOTO            CA1608S

                COMF            BARGB0, F
                INCF            BARGB0, F

CA1608S         BTFSS           AARGB0,MSB        ; if MSB set, negate AARG
                GOTO            C1608SX

                COMF            AARGB1, F
                COMF            AARGB0, F
                INCF            AARGB1, F
                BTFSC           _Z
                INCF            AARGB0, F

C1608SX		MOVF		AARGB0,W
		IORWF		BARGB0,W
		MOVWF		TEMP
		BTFSC		TEMP,MSB
		GOTO		C1608SX1

C1608S          SDIV1608

		BTFSC		TEMPB3,LSB		; test exception flag
		GOTO		C1608SX4

C1608SOK        BTFSS           SIGN,MSB
                RETLW           0x00

                COMF            AARGB1, F
                COMF            AARGB0, F
                INCF            AARGB1, F
                BTFSC           _Z
                INCF            AARGB0, F

                COMF            REMB0, F
                INCF            REMB0, F

                RETLW           0x00

C1608SX1	BTFSS		BARGB0,MSB		; test BARG exception
		GOTO		C1608SX3
		BTFSC		AARGB0,MSB		; test AARG exception
		GOTO		C1608SX2
		MOVF		AARGB1,W
		MOVWF		REMB0
		BCF		REMB0,MSB
		RLF		AARGB1,F
		RLF		AARGB0,F
		MOVF		AARGB0,W
		MOVWF		AARGB1
		CLRF		AARGB0
		GOTO		C1608SOK
C1608SX2	CLRF		AARGB1			; quotient = 1, remainder = 0
		INCF		AARGB1,F
		CLRF		AARGB0
		RETLW		0x00

C1608SX3	COMF		AARGB0,F		; numerator = 0x7FFF + 1
		COMF		AARGB1,F
		INCF		TEMPB3,F
		GOTO		C1608S

C1608SX4	INCF		REMB0,F			; increment remainder and test for
		MOVF		BARGB0,W		; overflow
		SUBWF		REMB0,W
		BTFSS		_Z
		GOTO		C1608SOK
		CLRF		REMB0			; if remainder overflow, clear
		INCF		AARGB1,F		; remainder, increment quotient and
		BTFSC		_Z
		INCF		AARGB0,F		; test for overflow exception
		BTFSS		AARGB0,MSB
		GOTO		C1608SOK
		BSF		FPFLAGS,NAN
		RETLW		0xFF

;**********************************************************************************************
;**********************************************************************************************
        
;       16/8 Bit Unsigned Fixed Point Divide 16/8 -> 16.08

;       Input:  16 bit unsigned fixed point dividend in AARGB0, AARGB1
;               8 bit unsigned fixed point divisor in BARGB0

;       Use:    CALL    FXD1608U

;       Output: 16 bit unsigned fixed point quotient in AARGB0, AARGB1
;               8 bit unsigned fixed point remainder in REMB0

;       Result: AARG, REM  <--  AARG / BARG

;       Max Timing:     1+291+2 = 294 clks

;       Min Timing:     1+227+2 = 230 clks

;       PM: 1+39+1 = 41         DM: 7

FXD1608U        CLRF            REMB0

                UDIV1608L

                RETLW           0x00

;**********************************************************************************************
;**********************************************************************************************
        
;       16/7 Bit Unsigned Fixed Point Divide 16/7 -> 16.07

;       Input:  16 bit unsigned fixed point dividend in AARGB0, AARGB1
;               7 bit unsigned fixed point divisor in BARGB0

;       Use:    CALL    FXD1607U

;       Output: 16 bit unsigned fixed point quotient in AARGB0, AARGB1
;               7 bit unsigned fixed point remainder in REMB0

;       Result: AARG, REM  <--  AARG / BARG

;       Max Timing:     1+171+2 = 174 clks

;       Min Timing:     1+171+2 = 174 clks

;       PM: 1+39+1 = 41         DM: 5

FXD1607U        CLRF            REMB0

                UDIV1607L

                RETLW           0x00

;**********************************************************************************************
;**********************************************************************************************
        
;       15/7 Bit Unsigned Fixed Point Divide 15/7 -> 15.07

;       Input:  15 bit unsigned fixed point dividend in AARGB0, AARGB1
;               7 bit unsigned fixed point divisor in BARGB0

;       Use:    CALL    FXD1507U

;       Output: 15 bit unsigned fixed point quotient in AARGB0, AARGB1
;               7 bit unsigned fixed point remainder in REMB0

;       Result: AARG, REM  <--  AARG / BARG

;       Max Timing:     1+163+2 = 166 clks

;       Min Timing:     1+163+2 = 166 clks

;       PM: 1+42+1 = 44         DM: 5

FXD1507U        CLRF            REMB0

                UDIV1507L

                RETLW           0x00

;**********************************************************************************************
;**********************************************************************************************

                END

	end
